#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX_NBR_OF_LINES	16384	/* maxi. # of lines/block */
#define MAX_NBR_OF_LABELS	1024	/* maxi. # of labels/block */
#define STACK_DEPTH             16      /* maxi. # of nested do-enddo pairs */

char *lines[MAX_NBR_OF_LINES];
long labels[MAX_NBR_OF_LINES];
long label_stack[STACK_DEPTH];

int verbose=0, upcase=0, label_step=10;

char *strupr(char *s);
char *strupr0(char *s);
int strnicmp(const char *s1, const char *s2, int n);

int main(argc,argv)
int  argc;
char *argv[];
{
  char s[256],s1[256],*p,*p1;
  register int i,j,n,m,gap,sp;
  long l;
  int nlines,nlabels,n_new;
  FILE *f_in;

  f_in=NULL;
  nlines=nlabels=0;

  p1=NULL;
  for (i=1;i<argc;i++) {
    p=argv[i];
    if (*p=='-') {
      if (!*(++p)) continue;
      switch (toupper(*p)) {
	case 'V' :   /* verbose mode */
	  verbose=1;
	  break;
	case 'U' :   /* verbose mode */
	  upcase=1;
	  break;
        default  :
          fprintf(stderr,"Invalid option \"%s\".\n",argv[i]);
    usage:
          fprintf(stderr,"Usage:\n");
          fprintf(stderr,"\tdo-enddo [-v] [-u] [inputfile] >outputfile\n");
          return 1;
      }
    } else {
      if (p1!=NULL) {
        printf("Too many file names!\n"); goto usage;
      }
      p1=argv[i];
    }
  }

  if (verbose)
    fprintf(stderr,"Conversion of DO-ENDDO pairs to conventional DO loops.\n");

  if (p1==NULL) f_in=stdin; else {
    if ((f_in=fopen(p1,"rt"))==NULL) {
      fprintf(stderr,"Cannot open input file \"%s\".\n",p1);
      return 1;
    }
  }

  /* reading input file */
  if (verbose) fprintf(stderr,"Reading input file: Line ");
  j=0;
  while (1) {
    p=s;
    do {
      if ((i=fgetc(f_in))==EOF) break;
      *p=i;
      if (*p=='\n') break;
      p++;
    } while (1);
    *p='\0';
    if (i==EOF && p==s) {
      if (verbose) fprintf(stderr,"%-5d",j); break;
    }
    if (!(++j & 15) && verbose) fprintf(stderr,"%-5d%c%c%c%c%c",j,8,8,8,8,8);
    if (nlines>=MAX_NBR_OF_LINES) {
      fprintf(stderr,"Too many lines!\n"); goto error;
    }
    n=strlen(s);
    if ((lines[nlines]=malloc(n+1))==NULL) {
      fprintf(stderr,"Memory overflow!\n"); goto error;
    }
    for (i=0;i<=n;i++) lines[nlines][i]=s[i];
    nlines++;
    /* check if a label is defined */
    for (i=0;i<5;i++) {
      if (s[i]==' ') continue;
      if (s[i]=='\t') break;
      if (s[i]<'0' || s[i]>'9') break;
      /* a label is defined */
      if (nlabels>=MAX_NBR_OF_LABELS) {
        fprintf(stderr,"Too many labels!\n"); goto error;
      }
      n=i+1;
      while (n<5 && s[n]>='0' && s[n]<='9') n++;
      s[n]='\0';
      labels[nlabels++]=atol(s+i);
      break;
    }
  }
  fclose(f_in); f_in=NULL;

  /* sorting labels */
  if (verbose) fprintf(stderr,"Sorting labels...");
  for (gap=nlabels/2; gap>0; gap>>=1) {
    for (i=gap; i<nlabels; i++) {
      for (j=i-gap; j>=0 && labels[j]>labels[j+gap]; j-=gap) {
        l=labels[j]; labels[j]=labels[j+gap]; labels[j+gap]=l;
      }
    }
  }
  if (verbose) {
    fprintf(stderr,"\n");
    fprintf(stderr,"Number of line(s) read:   %d\n",nlines);
    fprintf(stderr,"Number of label(s) found: %d\n",nlabels);
  }

  /* output new file */
  l=0l; gap=0; sp=0; n_new=0;	/* l=label */
  for (j=0;j<nlines;j++) {
    i=0;
    do {
      s[i]=lines[j][i];
      if (!s[i]) break;
      i++;
    } while (1);
    if (s[0]=='c' || s[0]=='C' || s[0]=='*') {  /* comment line */
      goto copy_line;
    }
    if (upcase) strupr0(s);
    /* skip leading blanks & label */
    for (p=s;p-s<6 && *p;p++) {
      if (*p=='\t') {p++; break;}
    }
    m=p-s;
    while (*p==' ' || *p=='\t') p++;
    n=p-s;  /* save position */
    /* cancel blank characters in the input line */
    p1=s1;
    do {
      if (!*p || p-s>=72) {*p1='\0'; break;}
      if (*p!=' ' && *p!='\t') *(p1++)=*p;
      p++;
    } while (1);

    if (!strnicmp(s1,"enddo",5) && !s1[5]) {
      if (!sp) {
        fprintf(stderr,"Warning: line %d, extra ENDDO copied to the output file\n",j+1);
        goto copy_line;
      } else {
        printf("%5ld ",label_stack[--sp]);
        for (i=m;i<n;i++) putchar(s[i]);
        printf("CONTINUE\n");
      }
    } else if (!strnicmp(s1,"do",2)) {
      /* seek for '=' */
      p1=s1+2;
      while (*p1!='=' && *p1) p1++;
      if (*p1!='=') {
/* may be "double precision..."        fprintf(stderr,"Warning: line %d, incomplete DO instruction.\n",j+1); */
        goto copy_line;
      }
      /* seek for ',' */
      p1++;
      while (*p1 && *p1!=',') {
        if (*p1=='(') {
          register int i,flag;
          flag=0; i=1; p1++;
          while (i && *p1) {
            if (*p1=='(') i++; else if (*p1==')') {
              if (!i) flag=1; else i--;
            }
            p1++;
          }
          if (i || flag) {
            if (i) fprintf(stderr,"Warning: line %d, missing right parenthesis.\n",j+1);
            if (flag) fprintf(stderr,"Warning: line %d, missing left parenthesis.\n",j+1);
            goto copy_line;
          }
        } else {
          if (*p1==')') {
            fprintf(stderr,"Warning: line %d, extra right parenthesis.\n",j+1);
          }
          p1++;
        }
      }
      if (*p1!=',') goto copy_line;	/* normal assignment */
      if (s1[2]>='0' && s1[2]<='9') {   /* normal DO instruction */
        printf("%s\n",s); continue;
      } else if (sp>=STACK_DEPTH) {
        fprintf(stderr,"Warning: line %d, too many nested do-enddo pairs.\n",j+1);
        goto copy_line;
      } else {
        /* find next unused label */
        l+=label_step;
        if (gap<nlabels) {
          do {
            while (gap<nlabels && l>labels[gap]) gap++;
            if (gap<nlabels) {
              if (l==labels[gap]) {l+=label_step; continue;}
            }
            break;
          } while (1);
        }
        label_stack[sp++]=l;
	strncpy(s1,s,n);
	sprintf(s1+n,"%c%c %ld%s",s[n],s[n+1],l,s+n+2);
	if (strlen(s1)<=72) printf("%s\n",s1); else {
	  strcpy(s,s1+72); s1[72]='\0';
	  printf("%s\n     &%s\n",s1,s);
	}
	n_new++;
      }
    } else { /* other instructions */
copy_line:
      printf("%s\n",s); continue;
    }
  }

  if (sp) {
    fprintf(stderr,"Warning: unbalanced DO-ENDDO instruction(s) found.\n");
  }

  if (verbose) fprintf(stderr,"%d new label(s) created.\n",n_new);

  if (f_in!=NULL && f_in!=stdin) {fclose(f_in); f_in=NULL;}
  return 0;

error:
  if (f_in!=NULL && f_in!=stdin) {fclose(f_in); f_in=NULL;}
  return 1;
}

char *strupr(char *s)
{
  register char *p;

  for (p=s; *p; p++) *p=toupper(*p);

  return s;
}

#define singlequote	'\''

char *strupr0(char *s)
/* change only characters which are not delimited by "'" to upcases */
{
  register char *p0, *p1;
  char c;

  for (p0=s; *p0; ) {
    for (p1=p0; *p1!=singlequote && *p1; p1++);
    c=*p1; *p1='\0'; strupr(p0); *p1=c;
    if (*p1!=singlequote) {p0=p1; break;}
    /* look for closing single quote */
    ++p1;
    while (*p1) {
      if (*p1==singlequote) {
        if (*(p1+1)==singlequote) {p1+=2; continue;}
        p1++;
        break;
      }
      p1++;
    }
    p0=p1;
  }

  return s;
}

int strnicmp(const char *s1, const char *s2, int n)
{
  char t1[1024], t2[1024];

  strncpy(t1, s1, n); t1[n]='\0';
  strncpy(t2, s2, n); t2[n]='\0';

  return strncmp(strupr(t1), strupr(t2), n);
}
